import logging
from functools import cached_property, lru_cache
from typing import Any, List, Optional, Union

from async_lru import alru_cache
from brownie import chain
from multicall import Call
from multicall.utils import await_awaitable
from y import convert
from y.classes.common import ERC20, ContractBase
from y.classes.singleton import Singleton
from y.datatypes import AddressOrContract, AnyAddressType, Block, UsdPrice
from y.networks import Network
from y.utils.logging import yLazyLogger
from y.utils.multicall import fetch_multicall
from y.utils.raw_calls import raw_call_async

logger = logging.getLogger(__name__)


v1_pools = {
    Network.Mainnet: [
        "0x398eC7346DcD622eDc5ae82352F02bE94C62d119"
    ],
}.get(chain.id, [])

v2_pools = {
    Network.Mainnet: [
        "0x7d2768dE32b0b80b7a3454c06BdAc94A69DDc7A9", # aave
        "0x7937D4799803FbBe595ed57278Bc4cA21f3bFfCB", # aave amm
        "0xcE744a9BAf573167B2CF138114BA32ed7De274Fa", # umee
    ],
    Network.Polygon: [
        "0x8dFf5E27EA6b7AC08EbFdf9eB090F32ee9a30fcf", # aave
    ],
    Network.Avalanche: [
        "0x70BbE4A294878a14CB3CDD9315f5EB490e346163", # blizz
    ]
}.get(chain.id, [])        


class AaveMarketBase(ContractBase):
    def __init__(self, address: AnyAddressType, *args: Any, **kwargs: Any) -> None:
        super().__init__(address, *args, **kwargs)
    
    def __contains__(self, __o: object) -> bool:
        return convert.to_address(__o) in self.atokens


class AaveMarketV1(AaveMarketBase):
    def __init__(self, address: AnyAddressType, *args: Any, **kwargs: Any) -> None:
        super().__init__(address, *args, **kwargs)
    
    def __repr__(self) -> str:
        return f"<AaveMarketV1 '{self.address}'>"
    
    #yLazyLogger(logger)
    @lru_cache
    def underlying(self, token_address: AddressOrContract) -> ERC20:
        return await_awaitable(self.underlying_async(token_address))
    
    #yLazyLogger(logger)
    @alru_cache
    async def underlying_async(self, token_address: AddressOrContract) -> ERC20:
        underlying = await raw_call_async(token_address, 'underlyingAssetAddress()',output='address')
        print(underlying)
        return ERC20(underlying)
    
    @cached_property
    #yLazyLogger(logger)
    def atokens(self) -> List[ERC20]:
        reserves_data = Call(self.address, ['getReserves()(address[])'], [[self.address,None]])()[self.address]
        reserves_data = fetch_multicall(*[[self.contract, 'getReserveData', reserve] for reserve in reserves_data])
        atokens = [ERC20(reserve['aTokenAddress']) for reserve in reserves_data]
        logger.info(f'loaded {len(atokens)} v1 atokens for {self.__repr__()}')
        return atokens


class AaveMarketV2(AaveMarketBase):
    def __init__(self, address: AnyAddressType, *args: Any, **kwargs: Any) -> None:
        super().__init__(address, *args, **kwargs)
    
    def __repr__(self) -> str:
        return f"<AaveMarketV2 '{self.address}'>"
    
    #yLazyLogger(logger)
    @lru_cache
    def underlying(self, token_address: AddressOrContract) -> ERC20:
        return await_awaitable(self.underlying_async(token_address))
    
    #yLazyLogger(logger)
    @alru_cache
    async def underlying_async(self, token_address: AddressOrContract) -> ERC20:
        underlying = await raw_call_async(token_address, 'UNDERLYING_ASSET_ADDRESS()',output='address')
        print(underlying)
        return ERC20(underlying)

    @cached_property
    #yLazyLogger(logger)
    def atokens(self) -> List[ERC20]:
        reserves = Call(self.address, ['getReservesList()(address[])'], [[self.address,None]])()[self.address]
        reserves_data = [
            Call(
                self.address,
                ['getReserveData(address)((uint256,uint128,uint128,uint128,uint128,uint128,uint40,address,address,address,address,uint8))',reserve],
            )()
            for reserve in reserves
        ]

        try:
            atokens = [ERC20(reserve_data[7]) for reserve_data in reserves_data]
            logger.info(f'loaded {len(atokens)} v2 atokens for {self.__repr__()}')
            return atokens
        except TypeError: # TODO figure out what to do about non verified aave markets
            logger.error(f'failed to load tokens for {self.__repr__()}')
            return []


class AaveRegistry(metaclass = Singleton):
    def __init__(self) -> None:
        pass

    @cached_property
    def pools(self) -> List[Union[AaveMarketV1, AaveMarketV2]]:
        return self.pools_v1 + self.pools_v2
    
    @cached_property
    def pools_v1(self) -> List[AaveMarketV1]:
        return [AaveMarketV1(pool) for pool in v1_pools]
    
    @cached_property
    def pools_v2(self) -> List[AaveMarketV2]:
        return [AaveMarketV2(pool) for pool in v2_pools]
    
    #yLazyLogger(logger)
    def pool_for_atoken(self, token_address: AnyAddressType) -> Optional[Union[AaveMarketV1, AaveMarketV2]]:
        for pool in self.pools:
            if token_address in pool:
                return pool

    def __contains__(self, __o: object) -> bool:
        return any(__o in pool for pool in self.pools)

    #yLazyLogger(logger)
    @lru_cache
    def is_atoken(self, token_address: AnyAddressType) -> bool:
        return token_address in self
    
    #yLazyLogger(logger)
    @lru_cache
    def underlying(self, token_address: AddressOrContract) -> ERC20:
        return await_awaitable(self.underlying_async(token_address))
    
    #yLazyLogger(logger)
    @alru_cache
    async def underlying_async(self, token_address: AddressOrContract) -> ERC20:
        pool = self.pool_for_atoken(token_address)
        return await pool.underlying_async(token_address)
    
    #yLazyLogger(logger)
    def get_price(self, token_address: AddressOrContract, block: Optional[Block] = None) -> UsdPrice:
        return await_awaitable(self.get_price_async(token_address, block=block))
    
    #yLazyLogger(logger)
    async def get_price_async(self, token_address: AddressOrContract, block: Optional[Block] = None) -> UsdPrice:
        underlying = await self.underlying_async(token_address)
        return await underlying.price_async(block)


aave = AaveRegistry()




import asyncio
import logging
from functools import cached_property, lru_cache
from typing import Any, Optional, Union

import brownie
from async_property import async_cached_property, async_property
from brownie.exceptions import ContractNotFound
from checksum_dict import ChecksumAddressSingletonMeta
from multicall.utils import await_awaitable, gather
from y import convert
from y.constants import EEE_ADDRESS
from y.contracts import (Contract, build_name, has_method, has_method_async,
                         probe)
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.erc20 import decimals_async, totalSupply_async
from y.exceptions import (ContractNotVerified, MessedUpBrownieContract,
                          NonStandardERC20)
from y.networks import Network
from y.prices import magic
from y.utils.logging import yLazyLogger
from y.utils.raw_calls import balanceOf_async, raw_call_async

logger = logging.getLogger(__name__)


class ContractBase(metaclass=ChecksumAddressSingletonMeta):
    def __init__(self, address: AnyAddressType, *args: Any, **kwargs: Any) -> None:
        self.address = convert.to_address(address)
        super().__init__(*args, **kwargs)
    
    def __str__(self) -> str:
        return f'{self.address}'
    
    def __eq__(self, __o: object) -> bool:
        try:
            return convert.to_address(__o) == self.address
        except:
            return False
    
    def __hash__(self) -> int:
        return hash(self.address)
    
    @cached_property
    def contract(self) -> brownie.Contract:
        return Contract(self.address)
    
    @cached_property
    #@yLazyLogger(logger)
    def _is_cached(self) -> bool:
        try:
            self.contract
            return True
        except (ContractNotVerified):
            return False
        except (ContractNotFound, MessedUpBrownieContract):
            return None
    
    @cached_property
    #@yLazyLogger(logger)
    def build_name(self) -> str:
        return build_name(self.address)
    
    #@yLazyLogger(logger)
    @lru_cache
    def has_method(self, method: str, return_response: bool = False) -> Union[bool,Any]:
        return has_method(self.address, method, return_response=return_response)
    
    async def has_method_async(self, method: str, return_response: bool = False) -> Union[bool,Any]:
        return await has_method_async(self.address, method, return_response=return_response)


class ERC20(ContractBase):
    def __init__(self, address: AnyAddressType, *args: Any, **kwargs: Any) -> None:
        super().__init__(address, *args, **kwargs)
    
    def __repr__(self) -> str:
        if asyncio.get_event_loop().is_running() or self.symbol is None:
            return f"<ERC20 '{self.address}'>"
        else:
            return f"<ERC20 {self.symbol} '{self.address}'>"
    
    @cached_property
    def symbol(self) -> str:
        return await_awaitable(self.symbol_async)

    @async_cached_property
    async def symbol_async(self) -> str:
        if self.address == EEE_ADDRESS:
            return "ETH"
        
        symbol = await probe(self.address, ["symbol()(string)", "SYMBOL()(string)", "getSymbol()(string)"])
        if symbol:
            return symbol

        # we've failed to fetch
        self.raise_exception('symbol')
    
    @cached_property
    def name(self) -> str:
        return await_awaitable(self.name_async)
    
    @async_cached_property
    async def name_async(self) -> str:
        if self.address == EEE_ADDRESS:
            return "Ethereum"
        
        name = await probe(self.address, ["name()(string)", "NAME()(string)", "getName()(string)"])
        if name:
            return name
        
        # we've failed to fetch
        self.raise_exception('name')
    
    #@yLazyLogger(logger)
    @async_cached_property
    async def decimals(self) -> int:
        if self.address == EEE_ADDRESS:
            return 18
        return await decimals_async(self.address)

    #@yLazyLogger(logger)
    async def _decimals(self, block: Optional[Block] = None) -> int:
        '''used to fetch decimals at specific block'''
        if self.address == EEE_ADDRESS:
            return 18
        return await decimals_async(self.address, block=block)
    
    #@yLazyLogger(logger)
    @async_cached_property
    async def scale(self) -> int:
        return 10 ** await self.decimals
    
    #@yLazyLogger(logger)
    async def _scale(self, block: Optional[Block] = None) -> int:
        return 10 ** await self._decimals(block=block)

    def total_supply(self, block: Optional[Block] = None) -> int:
        return await_awaitable(self.total_supply_async(block=block))

    #@yLazyLogger(logger)
    async def total_supply_async(self, block: Optional[Block] = None) -> int:
        return await totalSupply_async(self.address, block=block)
    
    def total_supply_readable(self, block: Optional[Block] = None) -> float:
        return await_awaitable(self.total_supply_readable_async(block=block))

    #@yLazyLogger(logger)
    async def total_supply_readable_async(self, block: Optional[Block] = None) -> float:
        total_supply, scale = await gather([self.total_supply_async(block=block), self.scale])
        return total_supply / scale
    
    def balance_of(self, address: AnyAddressType, block: Optional[Block] = None) -> int:
        return await_awaitable(self.balance_of_async(address, block=block))
    
    #@yLazyLogger(logger)
    async def balance_of_async(self, address: AnyAddressType, block: Optional[Block] = None) -> int:
        return await balanceOf_async(self.address, address, block=block)
    
    def balance_of_readable(self, address: AnyAddressType, block: Optional[Block] = None) -> float:
        return await_awaitable(self.balance_of_readable_async(address, block=block))
    
    #@yLazyLogger(logger)
    async def balance_of_readable_async(self, address: AnyAddressType, block: Optional[Block] = None) -> float:
        balance, scale = await gather([self.balance_of_async(address, block=block), self.scale])
        return balance / scale

    def price(self, block: Optional[Block] = None, return_None_on_failure: bool = False) -> Optional[UsdPrice]:
        return await_awaitable(
            self.price_async(block=block, return_None_on_failure=return_None_on_failure)
        )

    #yLazyLogger(logger)
    async def price_async(self, block: Optional[Block] = None, return_None_on_failure: bool = False) -> Optional[UsdPrice]:
        return await magic.get_price_async(
            self.address, 
            block=block, 
            fail_to_None=return_None_on_failure
        )
    
    def raise_exception(self, fn_name: str):
        raise NonStandardERC20(f'''
            Unable to fetch `{fn_name}` for {self.address} on {Network.printable()}
            If the contract is verified, please check to see if it has a strangely named
            `{fn_name}` method and create an issue on https://github.com/BobTheBuidler/ypricemagic
            with the contract address and correct method name so we can keep things going smoothly :)''')

class WeiBalance:
    def __init__(
        self, balance: int,
        token: AnyAddressType,
        block: Optional[Block] = None
        ) -> None:

        self.balance = balance
        self.token = ERC20(str(token))
        self.block = block
        super().__init__()

    def __str__(self) -> str:
        return str(self.balance)

    def __eq__(self, __o: object) -> bool:
        return __o == self.balance
    
    @property
    def readable(self) -> float:
        return await_awaitable(self.readable_async)
    
    @async_property
    async def readable_async(self) -> float:
        if self.balance == 0:
            return 0
        return self.balance / await self.token.scale

    @property
    def value_usd(self) -> float:
        return await_awaitable(self.value_usd_async)
    
    @async_property
    async def value_usd_async(self) -> float:
        if self.balance == 0:
            return 0
        balance, price = await gather([
            self.readable_async,
            self.token.price_async(block=self.block),
        ])
        return balance * price



import asyncio
import logging
from functools import cached_property, lru_cache
from typing import Any, Awaitable, Dict, List, Optional, Tuple

from async_lru import alru_cache
from brownie import chain
from brownie.convert.datatypes import EthAddress
from hexbytes import HexBytes
from multicall import Call
from multicall.utils import await_awaitable, gather
from y.classes.common import ERC20, ContractBase, WeiBalance
from y.classes.singleton import Singleton
from y.constants import STABLECOINS, WRAPPED_GAS_COIN
from y.contracts import build_name_async, has_methods_async
from y.datatypes import Address, AnyAddressType, Block, UsdPrice, UsdValue
from y.networks import Network
from y.utils.events import decode_logs, get_logs_asap_async
from y.utils.logging import yLazyLogger
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)

BALANCER_V2_VAULTS = {
    Network.Mainnet: [
        '0xBA12222222228d8Ba445958a75a0704d566BF2C8',
    ],
    Network.Fantom: [
        '0x20dd72Ed959b6147912C2e529F0a0C651c33c9ce',
    ],
    Network.Polygon: [
        '0xBA12222222228d8Ba445958a75a0704d566BF2C8',
    ],
    Network.Arbitrum: [
        '0xBA12222222228d8Ba445958a75a0704d566BF2C8',
    ],
}.get(chain.id, [])


class PoolId(bytes):
    def __init__(self, v: int) -> None:
        super().__init__()


class BalancerV2Vault(ContractBase):
    def __init__(self, address: AnyAddressType, *args: Any, **kwargs: Any) -> None:
        super().__init__(address, *args, **kwargs)
        if not self._is_cached:
            # we need the contract cached so we can decode logs correctly
            self.contract
    
    def get_pool_tokens(self, pool_id: int, block: Optional[Block] = None):
        return await_awaitable(self.get_pool_tokens_async(pool_id, block))
    
    #yLazyLogger(logger)
    async def get_pool_tokens_async(self, pool_id: int, block: Optional[Block] = None):
        return await self.contract.getPoolTokens.coroutine(pool_id, block_identifier = block)

    #yLazyLogger(logger)
    def list_pools(self, block: Optional[Block] = None) -> Dict[HexBytes,EthAddress]:
        return await_awaitable(self.list_pools_async(block=block))
    
    #yLazyLogger(logger)
    @alru_cache(maxsize=10)
    async def list_pools_async(self, block: Optional[Block] = None) -> Dict[HexBytes,EthAddress]:
        topics = ['0x3c13bc30b8e878c53fd2a36b679409c073afd75950be43d8858768e956fbc20e']
        #try:
        events = decode_logs(await get_logs_asap_async(self.address, topics, to_block=block))
        #except TypeError as e:
        #    if "Start must be less than or equal to stop" in str(e):
        #        return {}
        #    raise
        return {event['poolId'].hex():event['poolAddress'] for event in events}
    
    @lru_cache(maxsize=10)
    def get_pool_info(self, poolids: Tuple[HexBytes,...], block: Optional[Block] = None) -> List[Tuple]:
        return await_awaitable(self.get_pool_info_async(poolids,block=block))
    
    @alru_cache(maxsize=10)
    async def get_pool_info_async(self, poolids: Tuple[HexBytes,...], block: Optional[Block] = None) -> List[Tuple]:
        return await gather([
            self.contract.getPoolTokens.coroutine(poolId, block_identifier=block)
            for poolId in poolids
        ])

    #yLazyLogger(logger)
    def deepest_pool_for(self, token_address: Address, block: Optional[Block] = None) -> Tuple[Optional[EthAddress],int]:
        return await_awaitable(self.deepest_pool_for_async(token_address, block=block))
    
    #yLazyLogger(logger)
    async def deepest_pool_for_async(self, token_address: Address, block: Optional[Block] = None) -> Tuple[Optional[EthAddress],int]:
        pools = await self.list_pools_async(block=block)
        is_standard_pool = await asyncio.gather(*[_is_standard_pool(pool) for pool in pools.values()])
        poolids = (poolid for (poolid, pool), is_standard in zip(pools.items(), is_standard_pool) if is_standard)
        pools_info = await self.get_pool_info_async(poolids, block=block)
        pools_info = {self.list_pools(block=block)[poolid]: info for poolid, info in zip(poolids, pools_info) if str(info) != "((), (), 0)"}
        
        deepest_pool = {'pool': None, 'balance': 0}
        for pool, info in pools_info.items():
            num_tokens = len(info[0])
            pool_balances = {info[0][i]: info[1][i] for i in range(num_tokens)}
            pool_balance = [balance for token, balance in pool_balances.items() if token == token_address]
            if len(pool_balance) == 0:
                continue
            assert len(pool_balance) == 1
            pool_balance = pool_balance[0]
            if pool_balance > deepest_pool['balance']:
                deepest_pool = {'pool': pool, 'balance': pool_balance}

        return deepest_pool['pool'], deepest_pool['balance']


class BalancerV2Pool(ERC20):
    def __init__(self, pool_address: AnyAddressType) -> None:
        super().__init__(pool_address)

    @cached_property
    #yLazyLogger(logger)
    def id(self) -> PoolId:
        return Call(self.address, ['getPoolId()(bytes32)'], [['id',PoolId]])()['id']
    
    @cached_property
    #yLazyLogger(logger)
    def vault(self) -> BalancerV2Vault:
        vault = raw_call(self.address,'getVault()',output='address')
        return BalancerV2Vault(vault)

    def get_pool_price(self, block: Optional[Block] = None) -> UsdPrice:
        return await_awaitable(self.get_pool_price_async(block=block))
    
    #yLazyLogger(logger)
    async def get_pool_price_async(self, block: Optional[Block] = None) -> Awaitable[UsdPrice]:
        tvl, total_supply = await gather([
            self.get_tvl_async(block=block),
            self.total_supply_readable_async(block=block),
        ])
        return UsdPrice(tvl / total_supply)

    def get_tvl(self, block: Optional[Block] = None) -> UsdValue:
        return await_awaitable(self.get_tvl_async(block=block))
        
    #yLazyLogger(logger)
    async def get_tvl_async(self, block: Optional[Block] = None) -> Awaitable[UsdValue]:
        balances = await self.get_balances_async(block=block)
        return UsdValue(sum(await gather([
            balance.value_usd_async for balance in balances.values()
            if balance.token.address != self.address  # NOTE: to prevent an infinite loop for tokens that include themselves in the pool (e.g. bb-a-USDC)
        ])))

    def get_balances(self, block: Optional[Block] = None) -> Dict[ERC20, WeiBalance]:
        return await_awaitable(self.get_balances_async(block=block))

    #yLazyLogger(logger)
    async def get_balances_async(self, block: Optional[Block] = None) -> Dict[ERC20, WeiBalance]:
        return {token: balance for token, balance in (await self.tokens_async(block=block)).items()}

    #yLazyLogger(logger)
    def get_token_price(self, token_address: AnyAddressType, block: Optional[Block] = None) -> Optional[UsdPrice]:
        return await_awaitable(self.get_token_price_async(self, token_address, block=block))

    #yLazyLogger(logger)
    async def get_token_price_async(self, token_address: AnyAddressType, block: Optional[Block] = None) -> Optional[UsdPrice]:
        token_balances = self.get_balances(block=block)
        pool_token_info = list(zip(token_balances.keys(),token_balances.values(), self.weights(block=block)))
        for pool_token, balance, weight in pool_token_info:
            if pool_token == token_address:
                token_balance, token_weight = balance, weight

        for pool_token, balance, weight in pool_token_info:
            if pool_token in STABLECOINS:
                paired_token_balance, paired_token_weight = balance, weight
                break
            elif pool_token == WRAPPED_GAS_COIN:
                paired_token_balance, paired_token_weight = balance, weight
                break
            elif len(pool_token_info) == 2 and pool_token != token_address:
                paired_token_balance, paired_token_weight = balance, weight
                break

        try:
            token_value_in_pool = paired_token_balance.value_usd / paired_token_weight * token_weight
            return UsdPrice(token_value_in_pool / token_balance.readable)
        except UnboundLocalError:
            return None
    
    def tokens(self, block: Optional[Block] = None) -> Dict[ERC20, WeiBalance]:
        return await_awaitable(self.tokens_async(block=block))

    #yLazyLogger(logger)
    async def tokens_async(self, block: Optional[Block] = None) -> Dict[ERC20, WeiBalance]:
        tokens, balances, lastChangedBlock = await self.vault.get_pool_tokens_async(self.id, block=block)
        return {ERC20(token): WeiBalance(balance, token, block=block) for token, balance in zip(tokens, balances)}

    #yLazyLogger(logger)
    def weights(self, block: Optional[Block] = None) -> List[int]:
        try:
            return self.contract.getNormalizedWeights(block_identifier = block)
        except (AttributeError,ValueError):
            return [1 for _ in self.tokens(block=block).keys()]


#yLazyLogger(logger)
@alru_cache(maxsize=None)
async def _is_standard_pool(pool: EthAddress) -> bool:
    '''
    Returns `False` if `build_name(pool) in ['ConvergentCurvePool','MetaStablePool']`, else `True`
    '''
    
    # With `return_None_on_failure=True`, if `build_name(pool)` fails,
    # we can't know for sure that its a standard pool, but... it probably is.
    return await build_name_async(pool, return_None_on_failure=True) not in ['ConvergentCurvePool','MetaStablePool']


class BalancerV2(metaclass=Singleton):
    def __init__(self) -> None:
        self.vaults = [BalancerV2Vault(vault) for vault in BALANCER_V2_VAULTS]
    
    def __str__(self) -> str:
        return "BalancerV2()"

    #yLazyLogger(logger)
    def is_pool(self, token_address: AnyAddressType) -> bool:
        return await_awaitable(self.is_pool_async(token_address))

    #yLazyLogger(logger)
    async def is_pool_async(self, token_address: AnyAddressType) -> bool:
        methods = ('getPoolId()(bytes32)','getPausedState()((bool,uint,uint))','getSwapFeePercentage()(uint)')
        return await has_methods_async(token_address, methods)
    
    def get_pool_price(self, pool_address: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
        return await_awaitable(self.get_pool_price_async(pool_address, block=block))

    #yLazyLogger(logger)
    async def get_pool_price_async(self, pool_address: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
        return await BalancerV2Pool(pool_address).get_pool_price_async(block=block)

    #yLazyLogger(logger)
    def get_token_price(self, token_address: Address, block: Optional[Block] = None) -> UsdPrice:
        return await_awaitable(self.get_token_price_async(token_address, block=block))
    
    #yLazyLogger(logger)
    async def get_token_price_async(self, token_address: Address, block: Optional[Block] = None) -> UsdPrice:
        deepest_pool = await self.deepest_pool_for_async(token_address, block=block)
        if deepest_pool is None:
            return
        return await deepest_pool.get_token_price_async(token_address, block)
    
    #yLazyLogger(logger)
    def deepest_pool_for(self, token_address: Address, block: Optional[Block] = None) -> Optional[BalancerV2Pool]:
        return await_awaitable(self.deepest_pool_for_async(token_address, block=block))
    
    #yLazyLogger(logger)
    async def deepest_pool_for_async(self, token_address: Address, block: Optional[Block] = None) -> Optional[BalancerV2Pool]:
        deepest_pools = await gather([vault.deepest_pool_for_async(token_address, block=block) for vault in self.vaults])
        deepest_pools = {vault.address: deepest_pool for vault, deepest_pool in zip(self.vaults, deepest_pools) if deepest_pool is not None}
        deepest_pool_balance = max(pool_balance for pool_address, pool_balance in deepest_pools.values())
        for pool_address, pool_balance in deepest_pools.values():
            if pool_balance == deepest_pool_balance and pool_address:
                return BalancerV2Pool(pool_address)

balancer = BalancerV2()




import asyncio
import json
import logging
import threading
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import brownie
import eth_retry
from async_lru import alru_cache
from brownie import chain, web3
from brownie.exceptions import ContractNotFound
from brownie.network.contract import (_add_deployment, _ContractBase,
                                      _DeployedContractBase,
                                      _fetch_from_explorer, _resolve_address)
from brownie.typing import AccountsType
from checksum_dict import ChecksumAddressDict, ChecksumAddressSingletonMeta
from dank_mids.brownie_patch import patch_contract
from hexbytes import HexBytes
from multicall import Call
from multicall.utils import await_awaitable, gather

from y import convert
from y.datatypes import Address, AnyAddressType, Block
from y.exceptions import (ContractNotVerified, NodeNotSynced, call_reverted,
                          contract_not_verified)
from y.interfaces.ERC20 import ERC20ABI
from y.networks import Network
from y.utils.cache import memory
from y.utils.dank_mids import dank_w3
from y.utils.logging import yLazyLogger

logger = logging.getLogger(__name__)

contract_threads = ThreadPoolExecutor(16)


# cached Contract instance, saves about 20ms of init time
_contract_lock = threading.Lock()

# These tokens have trouble when resolving the implementation via the chain.
FORCE_IMPLEMENTATION = {
    Network.Mainnet: {
        "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48": "0xa2327a938Febf5FEC13baCFb16Ae10EcBc4cbDCF", # USDC as of 2022-08-10
    },
}.get(chain.id, {})


def Contract_erc20(address: AnyAddressType) -> "Contract":
    address = convert.to_address(address)
    return Contract.from_abi('ERC20',address,ERC20ABI)


def Contract_with_erc20_fallback(address: AnyAddressType) -> "Contract":
    if isinstance(address, Contract):
        return address
    address = convert.to_address(address)
    try:
        return Contract(address)
    except ContractNotVerified:
        return Contract_erc20(address)


@memory.cache()
#yLazyLogger(logger)
def contract_creation_block(address: AnyAddressType, when_no_history_return_0: bool = False) -> int:
    """
    Determine the block when a contract was created using binary search.
    NOTE Requires access to historical state. Doesn't account for CREATE2 or SELFDESTRUCT.
    """
    address = convert.to_address(address)
    logger.info("contract creation block %s", address)
    height = chain.height

    if height == 0:
        raise NodeNotSynced(f'''
            `chain.height` returns 0 on your node, which means it is not fully synced.
            You can only use this function on a fully synced node.''')

    lo, hi = 0, height
    while hi - lo > 1:
        mid = lo + (hi - lo) // 2
        try:
            if web3.eth.get_code(address, mid):
                hi = mid
            else:
                lo = mid
        except ValueError as e:
            if 'missing trie node' in str(e):
                logger.critical('missing trie node, `contract_creation_block` may output a higher block than actual. Please try again using an archive node.')
                if when_no_history_return_0:
                    return 0
            elif 'Server error: account aurora does not exist while viewing' in str(e):
                logger.critical(str(e))
                if when_no_history_return_0:
                    return 0
            elif 'No state available for block' in str(e):
                logger.critical(str(e))
                if when_no_history_return_0:
                    return 0
            else:
                raise
            lo = mid
    return hi if hi != height else None


class Contract(brownie.Contract, metaclass=ChecksumAddressSingletonMeta):
    _ChecksumAddressSingletonMeta__instances: ChecksumAddressDict["Contract"]

    @eth_retry.auto_retry
    def __init__(
        self, 
        address: AnyAddressType, 
        owner: Optional[AccountsType] = None, 
        require_success: bool = True, 
        ) -> None:
        
        address = str(address)
        if address.lower() == "0xeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee":
            raise ContractNotFound(f"{address} is not a contract.")

        with _contract_lock:
            # autofetch-sources: false
            # Try to fetch the contract from the local sqlite db.
            try:
                super().__init__(address, owner=owner)
                self.verified = True
            # If we don't already have the contract in the db, we'll try to fetch it from the explorer.
            except ValueError as e:
                try:                  
                    name, abi = _resolve_proxy(address)
                    build = {"abi": abi, "address": address, "contractName": name, "type": "contract"}
                    self.__init_from_abi__(build, owner=owner, persist=True)
                except (ContractNotFound, ContractNotVerified) as e:
                    if require_success:
                        raise
                    if type(e) == ContractNotVerified:
                        self._verified = False
                    else:
                        self._verified = None
        
        patch_contract(self, dank_w3)   # Patch the Contract with coroutines for each method.
        _squeeze(self)                  # Get rid of unnecessary memory-hog properties

    @classmethod
    def from_abi(
        cls,
        name: str,
        address: str,
        abi: List,
        owner: Optional[AccountsType] = None,
        persist: bool = True,
        ) -> "Contract":
        self = cls.__new__(cls)
        build = {"abi": abi, "address": _resolve_address(address), "contractName": name, "type": "contract"}
        self.__init_from_abi__(build, owner, persist)
        patch_contract(self, dank_w3)   # Patch the Contract with coroutines for each method.
        _squeeze(self)                  # Get rid of unnecessary memory-hog properties
        return self
    
    @classmethod
    async def coroutine(
        cls, 
        address: AnyAddressType, 
        ) -> "Contract":
        try:
            return Contract._ChecksumAddressSingletonMeta__instances[address]
        except KeyError:
            return await asyncio.get_event_loop().run_in_executor(contract_threads, Contract, address)
    
    def __init_from_abi__(self, build: Dict, owner: Optional[AccountsType] = None, persist: bool = True) -> None:
        _ContractBase.__init__(self, None, build, {})  # type: ignore
        _DeployedContractBase.__init__(self, build['address'], owner, None)
        if persist:
            _add_deployment(self)
        self.verified = True
        return self

    def has_method(self, method: str, return_response: bool = False) -> Union[bool,Any]:
        return has_method(self.address, method, return_response=return_response)

    def has_methods(
        self, 
        methods: List[str],
        _func: Union[any, all] = all
    ) -> bool:
        return await_awaitable(self.has_methods_async(methods, _func))

    async def has_methods_async(
        self, 
        methods: List[str],
        _func: Union[any, all] = all
    ) -> bool:
        return await has_methods_async(self.address, methods, _func)

    def build_name(self, return_None_on_failure: bool = False) -> Optional[str]:
        return await_awaitable(self.build_name_async(return_None_on_failure=return_None_on_failure))

    async def build_name_async(self, return_None_on_failure: bool = False) -> Optional[str]:
        return await build_name_async(self.address, return_None_on_failure=return_None_on_failure)
    
    def get_code(self, block: Optional[Block] = None) -> HexBytes:
        return web3.eth.get_code(self.address, block=block)


@memory.cache()
#yLazyLogger(logger)
def is_contract(address: AnyAddressType) -> bool:
    '''
    Checks to see if the input address is a contract. Returns `True` if:
    - The address is not and has never been a contract
    - The address used to be a contract but has self-destructed
    '''
    address = convert.to_address(address)
    return web3.eth.get_code(address) not in ['0x',b'']

def has_method(address: Address, method: str, return_response: bool = False) -> Union[bool,Any]:
    return await_awaitable(has_method_async(address, method, return_response=return_response))

#yLazyLogger(logger)
@alru_cache(maxsize=None)
#@memory.cache()
async def has_method_async(address: Address, method: str, return_response: bool = False) -> Union[bool,Any]:
    '''
    Checks to see if a contract has a `method` view method with no inputs.
    `return_response=True` will return `response` in bytes if `response` else `False`
    '''
    address = convert.to_address(address)
    try:
        response = await Call(address, [method], [['key', None]]).coroutine()
        response = response['key']
    except Exception as e:
        if call_reverted(e):
            return False
        raise
    
    if response is None:
        return False
    if return_response:
        return response
    return True

def has_methods(
    address: AnyAddressType, 
    methods: Tuple[str],
    _func: Callable = all # Union[any, all]
) -> bool:
    return await_awaitable(has_methods_async(address,methods,_func=_func))

#yLazyLogger(logger)
@alru_cache(maxsize=None)
async def has_methods_async(
    address: AnyAddressType, 
    methods: Tuple[str],
    _func: Callable = all # Union[any, all]
) -> bool:
    '''
    Checks to see if a contract has each view method (with no inputs) in `methods`.
    Pass `at_least_one=True` to only verify a contract has at least one of the methods.
    '''

    assert _func in [all, any], '`_func` must be either `any` or `all`'

    address = convert.to_address(address)
    try:
        return _func([
            False if call is None else True
            for call
            in await gather([Call(address, [method]).coroutine() for method in methods])
        ])
    except Exception as e:
        if not call_reverted(e): raise # and not out_of_gas(e): raise
        # Out of gas error implies one or more method is state-changing.
        # If `_func == all` we return False because `has_methods` is only supposed to work for public view methods with no inputs
        # If `_func == any` maybe one of the methods will work without "out of gas" error
        return False if _func == all else any(await gather([has_method_async(address, method) for method in methods]))


#yLazyLogger(logger)
async def probe(
    address: AnyAddressType, 
    methods: List[str],
    block: Optional[Block] = None,
    return_method: bool = False
) -> Any:
    address = convert.to_address(address)
    results = await asyncio.gather(*[Call(address, [method], block_id=block).coroutine() for method in methods], return_exceptions=True)
    results = [(method, result) for method, result in zip(methods, results) if not isinstance(result, Exception) and result is not None]
    if len(results) not in [1,0]:
        if len(results) == 2 and results[0][1] == results[1][1]:
            method = results[0][0], results[1][0]
            result = results[0][1]
        else:
            raise AssertionError(f'`probe` returned multiple results for {address}: {results}. Must debug')
    if len(results) == 1:
        method, result = results[0]
    else:
        method, result = None, None
    
    if method:
        assert result is not None

    if not return_method:
        return result
    else:
        return method, result
    


@memory.cache()
#yLazyLogger(logger)
def build_name(address: AnyAddressType, return_None_on_failure: bool = False) -> str:
    return await_awaitable(build_name_async(address, return_None_on_failure=return_None_on_failure))

async def build_name_async(address: AnyAddressType, return_None_on_failure: bool = False) -> str:
    try:
        contract = await Contract.coroutine(address)
        return contract.__dict__['_build']['contractName']
    except ContractNotVerified:
        if not return_None_on_failure:
            raise
        return None

async def proxy_implementation(address: AnyAddressType, block: Optional[Block]) -> Address:
    return await probe(address, ['implementation()(address)','target()(address)'], block)

def _squeeze(it):
    """ Reduce the contract size in RAM significantly. """
    for k in ["ast", "bytecode", "coverageMap", "deployedBytecode", "deployedSourceMap", "natspec", "opcodes", "pcMap"]:
        if it._build and k in it._build.keys():
            it._build[k] = {}
    return it


def _extract_abi_data(address):
    try:
        data = _fetch_from_explorer(address, "getsourcecode", False)
    except ConnectionError as e:
        if '{"message":"Something went wrong.","result":null,"status":"0"}' in str(e):
            if chain.id == Network.xDai:
                raise ValueError(f'Rate limited by Blockscout. Please try again.')
            if web3.eth.get_code(address):
                raise ContractNotVerified(address)
            else:
                raise ContractNotFound(address)
        raise
    except ValueError as e:
        if contract_not_verified(e):
            raise ContractNotVerified(f'{address} on {Network.printable()}')
        elif "Unknown contract address:" in str(e):
            if not is_contract(address):
                raise ContractNotFound(str(e))
            raise ContractNotVerified(str(e)) # avax snowtrace
        else:
            raise
    
    is_verified = bool(data["result"][0].get("SourceCode"))
    if not is_verified:
        raise ContractNotVerified(f"Contract source code not verified: {address}")
    name = data["result"][0]["ContractName"]
    abi = json.loads(data["result"][0]["ABI"])
    implementation = data["result"][0]["Implementation"]
    return name, abi, implementation

@eth_retry.auto_retry
def _resolve_proxy(address) -> Tuple[str, List]:
    address = convert.to_address(address)
    name, abi, implementation = _extract_abi_data(address)
    as_proxy_for = None

    if address in FORCE_IMPLEMENTATION:
        implementation = FORCE_IMPLEMENTATION[address]
        name, abi, _ = _extract_abi_data(implementation)
        return name, abi

    # always check for an EIP1967 proxy - https://eips.ethereum.org/EIPS/eip-1967
    implementation_eip1967 = web3.eth.get_storage_at(
        address, int(web3.keccak(text="eip1967.proxy.implementation").hex(), 16) - 1
    )
    # always check for an EIP1822 proxy - https://eips.ethereum.org/EIPS/eip-1822
    implementation_eip1822 = web3.eth.get_storage_at(address, web3.keccak(text="PROXIABLE"))

    # Just leave this code where it is for a helpful debugger as needed.
    if address == "":
        raise Exception(
            f"""implementation: {implementation}
            implementation_eip1967: {len(implementation_eip1967)} {implementation_eip1967}
            implementation_eip1822: {len(implementation_eip1822)} {implementation_eip1822}""")

    if len(implementation_eip1967) > 0 and int(implementation_eip1967.hex(), 16):
        as_proxy_for = _resolve_address(implementation_eip1967[-20:])
    elif len(implementation_eip1822) > 0 and int(implementation_eip1822.hex(), 16):
        as_proxy_for = _resolve_address(implementation_eip1822[-20:])
    elif implementation:
        # for other proxy patterns, we only check if etherscan indicates
        # the contract is a proxy. otherwise we could have a false positive
        # if there is an `implementation` method on a regular contract.
        try:
            # first try to call `implementation` per EIP897
            # https://eips.ethereum.org/EIPS/eip-897
            c = Contract.from_abi(name, address, abi)
            as_proxy_for = c.implementation.call()
        except Exception:
            # if that fails, fall back to the address provided by etherscan
            as_proxy_for = _resolve_address(implementation)

    if as_proxy_for:
        name, abi, _ = _extract_abi_data(as_proxy_for)
    return name, abi




import logging
from typing import Optional

from brownie import ZERO_ADDRESS, chain
from brownie.exceptions import ContractNotFound
from multicall.utils import await_awaitable
from y.constants import usdc
from y.contracts import Contract
from y.datatypes import Address, Block, UsdPrice
from y.exceptions import ContractNotVerified, UnsupportedNetwork
from y.networks import Network
from y.utils.logging import yLazyLogger
from y.utils.raw_calls import _decimals

logger = logging.getLogger(__name__)

V1 = {
    Network.Mainnet: "0xc0a47dFe034B400B47bDaD5FecDa2621de6c4d95",
}.get(chain.id,None)

class UniswapV1:
    def __init__(self) -> None:
        self.factory = V1
    
    #yLazyLogger(logger)
    def get_price(self, token_address: Address, block: Optional[Block]) -> Optional[UsdPrice]:
        return await_awaitable(self.get_price_async(token_address, block=block))

    #yLazyLogger(logger)
    async def get_price_async(self, token_address: Address, block: Optional[Block]) -> Optional[UsdPrice]:
        try:
            factory = await Contract.coroutine(self.factory)
        except ValueError:
            raise UnsupportedNetwork
        exchange = factory.getExchange(token_address)
        if exchange == ZERO_ADDRESS:
            return None
        try:
            exchange = await Contract.coroutine(exchange)
            eth_bought = exchange.getTokenToEthInputPrice(10 ** await _decimals(token_address,block), block_identifier=block)
            exchange = await Contract.coroutine(factory.getExchange(usdc))
            usdc_bought = exchange.getEthToTokenInputPrice(eth_bought, block_identifier=block) / 1e6
            fees = 0.997 ** 2
            return UsdPrice(usdc_bought / fees)
        except (ContractNotFound, ContractNotVerified):
            pass
        except ValueError as e:
            if 'invalid jump destination' in str(e):
                return
            else:
                raise




import threading
from typing import Any, Dict, Optional, Tuple, Type, TypeVar

T = TypeVar("T", bound=object)


class Singleton(type):
    def __init__(cls, name: str, bases: Tuple[type, ...], namespace: Dict[str, Any]) -> None:
        cls.__instance: Optional[T] = None
        cls.__lock = threading.Lock()
        super().__init__(name, bases, namespace)

    def __call__(cls, *args: Any, **kwargs: Any) -> T:
        if cls.__instance is None:
            with cls.__lock:
                # Check again in case `__instance` was set while we were waiting for the lock.
                if cls.__instance is None:
                    cls.__instance = super().__call__(*args, **kwargs)
        return cls.__instance



import datetime
import logging

from brownie import chain, web3

from y.exceptions import NoBlockFound
from y.utils.cache import memory
from y.utils.client import get_ethereum_client
from y.utils.logging import yLazyLogger

logger = logging.getLogger(__name__)


@memory.cache() 
#yLazyLogger(logger)
def get_block_timestamp(height: int) -> int:
    client = get_ethereum_client()
    if client in ['tg', 'erigon']:
        header = web3.manager.request_blocking(f"{client}_getHeaderByNumber", [height])
        return int(header.timestamp, 16)
    else:
        return chain[height].timestamp


@memory.cache()
#yLazyLogger(logger)
def last_block_on_date(date_string: str) -> int:
    logger.debug(f'last block on date {date_string}')
    date = datetime.datetime.strptime(date_string, "%Y-%m-%d")
    date = date.date()
    height = chain.height
    lo, hi = 0, height
    while hi - lo > 1:
        mid = lo + (hi - lo) // 2
        logger.debug('block: ' + str(mid))
        logger.debug('mid: ' + str(datetime.date.fromtimestamp(get_block_timestamp(mid))))
        logger.debug(date)
        if datetime.date.fromtimestamp(get_block_timestamp(mid)) > date:
            hi = mid
        else:
            lo = mid
    hi = hi - 1
    return hi if hi != height else None


#yLazyLogger(logger)
def closest_block_after_timestamp(timestamp: int) -> int:
    try:
        return _closest_block_after_timestamp_cached(timestamp)
    except NoBlockFound:
        return None


@memory.cache()
def _closest_block_after_timestamp_cached(timestamp: int) -> int:
    logger.info('closest block after timestamp %d', timestamp)
    height = chain.height
    lo, hi = 0, height
    while hi - lo > 1:
        mid = lo + (hi - lo) // 2
        if get_block_timestamp(mid) > timestamp:
            hi = mid
        else:
            lo = mid
    if hi == height:
        raise NoBlockFound(f"No block found after timestamp {timestamp}")
    return hi
